#!/usr/bin/env python
# -*- coding: utf-8

# Copyright 2017-2019 The FIAAS Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime, timedelta, MINYEAR
import os.path


"""Singleton configuration for k8s client"""

#: API server URL
api_server = "https://kubernetes.default.svc.cluster.local"
#: API token
api_token = ""
# Used by in_cluster_configuration. Takes precedence over `api_token` and `cert`.
api_token_source = None
#: API certificate
cert = None
#: Should the client verify the servers SSL certificates?
verify_ssl = True
#: Enable debugging
debug = False
#: Default timeout for most operations
timeout = 20
#: Default timeout for streaming operations, used while waiting for more events.
#: When reached, the library will usually info log and reconnect.
#: There's a few considerations when setting this value:
#: * On some servers, it might take this long to detect a dropped connection. This speaks for a low value,
#:   to detect the issue faster.
#: * When connecting, a resourceVersion is used to resume, if still valid. This speaks for a low value,
#:   to avoid them expiring.
#: * During idle periods, there might not be any new resourceVersions.
#:   Bookmarks events are used to avoid this, they are sent at a server
#:   specific interval, but usually about once per minute.
#:   This speaks for a high value.
#: 4.5 minutes is the default, set to detect the first case above in a reasonable time,
#:while being just below the default resourceVersion expiration of 5 minutes.
stream_timeout = 270
#: Default size of Watcher cache. If you expect a lot of events, you might want to increase this.
watcher_cache_size = 1000


# disables bandit warning for this line which triggers because the string contains 'token', which is fine
def use_in_cluster_config(
    token_file="/var/run/secrets/kubernetes.io/serviceaccount/token",  # nosec
    ca_cert_file="/var/run/secrets/kubernetes.io/serviceaccount/ca.crt",
):
    """
    Configure the client using the recommended configuration for accessing the API from within a Kubernetes cluster:
    https://kubernetes.io/docs/tasks/access-application-cluster/access-cluster/#accessing-the-api-from-a-pod
    """
    global api_token_source
    global verify_ssl
    api_token_source = FileTokenSource(token_file)
    if os.path.exists(ca_cert_file):
        verify_ssl = ca_cert_file


class FileTokenSource(object):
    """Read API token from token_file, exposing it via token(). Calls to token() will re-read the token from file if
    more than 1 minute has passed since the last read.

    Intended to support the BoundServiceAccountTokenVolume feature in Kubernetes 1.21 and later.
    """

    def __init__(self, token_file, now_func=datetime.now):
        self._token_file = token_file
        self._expires_at = datetime(MINYEAR, 1, 1)  # force read on initial call to _refresh_token
        self._refresh_interval = timedelta(minutes=1)
        self._token = self._refresh_token(now_func=now_func)  # fail on init if token_file can not be read

    def token(self, now_func=datetime.now):
        return self._refresh_token(now_func=now_func)

    def _refresh_token(self, now_func=datetime.now):
        now = now_func()
        if self._expires_at <= now:
            with open(self._token_file, 'r') as f:
                self._token = f.read().strip()
                self._expires_at = now + self._refresh_interval
        return self._token



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2017-2019 The FIAAS Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import mock
import pytest

from k8s import config
from k8s.base import Model, WatchEvent
from k8s.fields import Field, RequiredField
from k8s.client import Client, SENSITIVE_HEADERS, _session_factory

import requests


# pylint: disable=R0201
@pytest.mark.usefixtures("k8s_config")
class TestClient(object):
    @pytest.fixture
    def success_response(self):
        resp = mock.MagicMock()
        resp.status_code = 200
        return resp

    @pytest.fixture
    def session(self, success_response):
        with mock.patch("k8s.client.Client._session") as mockk:
            mockk.request.return_value = success_response
            yield mockk

    @pytest.fixture
    def client(self):
        return Client()

    @pytest.fixture
    def url(self):
        return "/api/v1/nodes"

    @pytest.fixture
    def explicit_timeout(self):
        return 60

    @pytest.mark.parametrize("url", ["http://api.k8s.example.com", "https://api.k8s.example.com"])
    def test_session_configured_for_retry(self, url):
        session = _session_factory()
        adapter = session.get_adapter(url)
        assert adapter.max_retries.total > 0
        assert requests.codes.too_many_requests in adapter.max_retries.status_forcelist
        assert requests.codes.ok not in adapter.max_retries.status_forcelist

    def test_get_should_use_default_timeout(self, session, client, url):
        client.get(url)
        session.request.assert_called_once_with("GET", _absolute_url(url), json=None, timeout=config.timeout)

    def test_get_should_propagate_timeout(self, session, client, url, explicit_timeout):
        client.get(url, timeout=explicit_timeout)
        session.request.assert_called_once_with("GET", _absolute_url(url), json=None, timeout=explicit_timeout)

    def test_delete_should_use_default_timeout(self, session, client, url):
        client.delete(url)
        session.request.assert_called_once_with(
            "DELETE", _absolute_url(url), json=None, timeout=config.timeout
        )

    def test_delete_should_propagate_timeout(self, session, client, url, explicit_timeout):
        client.delete(url, timeout=explicit_timeout)
        session.request.assert_called_once_with("DELETE", _absolute_url(url), json=None, timeout=explicit_timeout)

    def test_delete_should_use_default_body(self, session, client, url):
        client.delete(url)
        session.request.assert_called_once_with(
            "DELETE", _absolute_url(url), json=None, timeout=config.timeout
        )

    def test_delete_should_propagate_body(self, session, client, url):
        body = {"kind": "DeleteOptions", "apiVersion": "v1", "propagationPolicy": "Foreground"}
        client.delete(url, body=body)
        session.request.assert_called_once_with(
            "DELETE", _absolute_url(url), json=body, timeout=config.timeout
        )

    def test_post_should_use_default_timeout(self, session, client, url):
        body = {"foo": "bar"}
        client.post(url, body=body)
        session.request.assert_called_once_with("POST", _absolute_url(url), json=body, timeout=config.timeout)

    def test_post_should_propagate_timeout(self, session, client, url, explicit_timeout):
        body = {"foo": "bar"}
        client.post(url, body=body, timeout=explicit_timeout)
        session.request.assert_called_once_with("POST", _absolute_url(url), json=body, timeout=explicit_timeout)

    def test_put_should_use_default_timeout(self, session, client, url):
        body = {"foo": "bar"}
        client.put(url, body=body)
        session.request.assert_called_once_with("PUT", _absolute_url(url), json=body, timeout=config.timeout)

    def test_put_should_propagate_timeout(self, session, client, url, explicit_timeout):
        body = {"foo": "bar"}
        client.put(url, body=body, timeout=explicit_timeout)
        session.request.assert_called_once_with("PUT", _absolute_url(url), json=body, timeout=explicit_timeout)

    def test_watch_list_should_raise_exception_when_watch_list_url_is_not_set_on_metaclass(self, session):
        with pytest.raises(NotImplementedError):
            list(WatchListExampleUnsupported.watch_list())

    def test_watch_list_with_namespace_should_raise_exception_when_watch_list_url_template_is_not_set(self, session):
        with pytest.raises(NotImplementedError):
            list(WatchListExampleUnsupported.watch_list(namespace="explicitly-set"))

    def test_watch_list(self, session):
        list(WatchListExample.watch_list())
        session.request.assert_called_once_with(
            "GET", _absolute_url("/watch/example"), json=None, timeout=config.stream_timeout, stream=True, params={}
        )

    def test_watch_list_with_namespace(self, session):
        list(WatchListExample.watch_list(namespace="explicitly-set"))
        session.request.assert_called_once_with(
            "GET",
            _absolute_url("/watch/explicitly-set/example"),
            json=None,
            timeout=config.stream_timeout,
            stream=True,
            params={},
        )

    def test_list_without_namespace_should_raise_exception_when_list_url_is_not_set_on_metaclass(self, session):
        with pytest.raises(NotImplementedError):
            WatchListExampleUnsupported.list(namespace=None)

    def test_list_default_namespace(self, session):
        WatchListExample.list()
        session.request.assert_called_once_with(
            "GET", _absolute_url("/apis/namespaces/default/example"), json=None, timeout=config.timeout
        )

    def test_list_explicit_namespace(self, session):
        WatchListExample.list(namespace="explicitly-set")
        session.request.assert_called_once_with(
            "GET", _absolute_url("/apis/namespaces/explicitly-set/example"), json=None, timeout=config.timeout
        )

    def test_list_without_namespace(self, session):
        WatchListExample.list(namespace=None)
        session.request.assert_called_once_with(
            "GET", _absolute_url("/example/list"), json=None, timeout=config.timeout
        )

    def test_find_without_namespace_should_raise_exception_when_list_url_is_not_set_on_metaclass(self, session):
        with pytest.raises(NotImplementedError):
            list(WatchListExampleUnsupported.find("foo", namespace=None))

    def test_find_default_namespace(self, session):
        WatchListExample.find("foo")
        session.request.assert_called_once_with(
            "GET", _absolute_url("/apis/namespaces/default/example"), json=None, timeout=config.timeout,
            params={"labelSelector": "app=foo"}
        )

    def test_find_explicit_namespace(self, session):
        WatchListExample.find("foo", namespace="explicitly-set")
        session.request.assert_called_once_with(
            "GET", _absolute_url("/apis/namespaces/explicitly-set/example"), json=None, timeout=config.timeout,
            params={"labelSelector": "app=foo"}
        )

    def test_find_without_namespace(self, session):
        WatchListExample.find("foo", namespace=None)
        session.request.assert_called_once_with(
            "GET", _absolute_url("/example/list"), json=None, timeout=config.timeout,
            params={"labelSelector": "app=foo"}
        )

    @pytest.mark.parametrize("key", SENSITIVE_HEADERS)
    def test_redacts_sensitive_headers(self, key):
        message = []
        sensitive_value = "super sensitive data that should not be exposed"
        Client._add_headers(message, {key: sensitive_value}, "")
        text = "".join(message)
        assert sensitive_value not in text


# pylint: disable=R0201
@pytest.mark.usefixtures("k8s_config")
class TestWatchListEvents(object):

    def test_watch_list_payload_ok(self, get):
        """
        verify watch events of WatchListExample create WatchEvent with the appropriate type and object
        """
        response = mock.create_autospec(requests.Response)
        response.status_code = 200
        response.iter_lines.return_value = ['''
{
  "type": "ADDED",
  "object": {
    "value": 1,
    "requiredValue": 2
  }
}''', '''
{
  "type": "MODIFIED",
  "object": {
    "value": 3,
    "requiredValue": 4
  }
}
''']
        get.return_value = response

        expected = [
            _create_watchevent(WatchEvent.ADDED, WatchListExample(value=1, requiredValue=2)),
            _create_watchevent(WatchEvent.MODIFIED, WatchListExample(value=3, requiredValue=4)),
        ]

        items = list(WatchListExample.watch_list())
        assert items == expected

    def test_watch_list_payload_invalid_json(self, get):
        """
        verify event which does not cleanly unmarshal from json to dict is discarded
        """
        response = mock.create_autospec(requests.Response)
        response.status_code = 200
        response.iter_lines.return_value = ['''
{
  "type": "ADDED",
  "object": {
    "value": 1,
    "requiredValue": 2
  }
}
''', '''
definitely not valid json
''', '''
{
  "type": "ADDED",
  "object": {
    "value": 5,
    "requiredValue": 6
  }
}''']
        get.return_value = response

        expected = [
            _create_watchevent(WatchEvent.ADDED, WatchListExample(value=1, requiredValue=2)),
            # "definitely not valid json" should be discarded
            _create_watchevent(WatchEvent.ADDED, WatchListExample(value=5, requiredValue=6)),
        ]

        items = list(WatchListExample.watch_list())
        assert items == expected

    def test_watch_list_payload_invalid_object(self, get):
        """
        verify event which contains a resource not valid according to the Model class is discarded
        """
        response = mock.create_autospec(requests.Response)
        response.status_code = 200
        response.iter_lines.return_value = ['''
{
  "type": "ADDED",
  "object": {
    "value": 1,
    "requiredValue": 2
  }
}
''', '''
{
  "type": "ADDED",
  "object": {
    "value": 10,
  }
}
''', '''
{
  "type": "ADDED",
  "object": {
    "value": 5,
    "requiredValue": 6
  }
}''']
        get.return_value = response

        expected = [
            _create_watchevent(WatchEvent.ADDED, WatchListExample(value=1, requiredValue=2)),
            # event with value=10 and requiredValue missing should be discarded
            _create_watchevent(WatchEvent.ADDED, WatchListExample(value=5, requiredValue=6)),
        ]

        items = list(WatchListExample.watch_list())
        assert items == expected


def _create_watchevent(event_type, event_object):
    """factory function for WatchEvent to make it easier to create test data from actual objects, as the constructor
    takes a dict (unmarshaled json)"""
    return WatchEvent({"type": event_type, "object": event_object.as_dict()}, event_object.__class__)


def _absolute_url(url):
    return config.api_server + url


class WatchListExample(Model):
    class Meta:
        list_url = "/example/list"
        url_template = "/apis/namespaces/{namespace}/example"
        watch_list_url = "/watch/example"
        watch_list_url_template = "/watch/{namespace}/example"

    value = Field(int)
    requiredValue = RequiredField(int)


class WatchListExampleUnsupported(Model):
    class Meta:
        url_template = "/example"

    value = Field(int)



#!/usr/bin/env python
# -*- coding: utf-8

# Copyright 2017-2019 The FIAAS Authors
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#      http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging

import mock
import pytest

from k8s import config


@pytest.fixture
def logger():
    """Set root logger to DEBUG, and add stream handler"""
    root_logger = logging.getLogger()
    old_level = root_logger.getEffectiveLevel()
    root_logger.setLevel(logging.DEBUG)
    handler = logging.StreamHandler()
    root_logger.addHandler(handler)
    yield root_logger
    root_logger.removeHandler(handler)
    root_logger.setLevel(old_level)


@pytest.fixture
def k8s_config(monkeypatch):
    """Configure k8s for test-runs"""
    monkeypatch.setattr(config, "api_server", "http://localhost:8080")
    monkeypatch.setattr(config, "verify_ssl", False)


@pytest.fixture
def post():
    with mock.patch('k8s.client.Client.post') as m:
        yield m


@pytest.fixture
def put():
    with mock.patch('k8s.client.Client.put') as m:
        yield m


@pytest.fixture
def get():
    with mock.patch('k8s.client.Client.get') as m:
        yield m


@pytest.fixture
def delete():
    with mock.patch('k8s.client.Client.delete') as m:
        yield m


@pytest.fixture
def api_get():
    with mock.patch('k8s.base.ApiMixIn.get') as m:
        yield m



#!/usr/bin/env python
# -*- coding: utf-8

# Copyright 2017-2019 The FIAAS Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from abc import ABC

import json
import logging
from collections import namedtuple
from typing import Optional

import requests
import requests.packages.urllib3 as urllib3

from . import config
from .client import Client, NotFound
from .fields import Field

LOG = logging.getLogger(__name__)
LOG.addHandler(logging.NullHandler())


class MetaModel(type):
    """Metaclass for Model

    Responsibilities:

    * Creating the _meta attribute, with url_template (if present), list of fields
      and for convenience, a list of field names.
    * Creates properties for name and namespace if the instance has a metadata field.
    * Mixes in ApiMixIn if the Model has a Meta attribute, indicating a top level
      Model (not to be confused with _meta).
    """

    @staticmethod
    def __new__(mcs, cls, bases, attrs):
        attr_meta = attrs.pop("Meta", None)
        if attr_meta:
            bases += (ApiMixIn,)
        meta = {
            "url_template": getattr(attr_meta, "url_template", ""),
            "list_url": getattr(attr_meta, "list_url", ""),
            "watch_list_url": getattr(attr_meta, "watch_list_url", ""),
            "watch_list_url_template": getattr(attr_meta, "watch_list_url_template", ""),
            "fields": [],
            "field_names": [],
        }
        field_names = meta["field_names"]
        fields = meta["fields"]
        for k, v in list(attrs.items()):
            if isinstance(v, Field):
                if v.name == "__unset__":
                    v.name = k
                v.attr_name = k
                field_names.append(k)
                fields.append(v)
        Meta = namedtuple("Meta", meta.keys())
        attrs["_meta"] = Meta(**meta)
        return super(MetaModel, mcs).__new__(mcs, cls, bases, attrs)


class ApiMixIn(object):
    """ApiMixIn class for top level Models

    Contains methods for working with the API
    """

    _client = Client()

    @classmethod
    def _build_url(cls, **kwargs):
        return cls._meta.url_template.format(**kwargs)

    @classmethod
    def find(cls, name="", namespace="default", labels=None):
        """Find resources using label selection

        :param: :py:class:`str` name: The name of the resource, as indicated by the `app` label
        :param: :py:class:`str` namespace: The namespace to search in
        :param: :py:class:`dict` labels: More advanced label selection (see below)

        :returns: A list of matching objects

        When a `labels` dictionary is supplied, the `name` parameter is ignored.
        See the docs for _label_selector for more details
        """
        if namespace is None:
            if not cls._meta.list_url:
                raise NotImplementedError("Cannot find without namespace, no list_url defined on class {}".format(cls))
            url = cls._meta.list_url
        else:
            url = cls._build_url(name="", namespace=namespace)
        if not labels:
            labels = {"app": Equality(name)}
        selector = cls._label_selector(labels)
        resp = cls._client.get(url, params={"labelSelector": selector})
        return [cls.from_dict(item) for item in resp.json()["items"]]

    @classmethod
    def list(cls, namespace="default"):
        """List all resources in given namespace"""
        if namespace is None:
            if not cls._meta.list_url:
                raise NotImplementedError("Cannot list without namespace, no list_url defined on class {}".format(cls))
            url = cls._meta.list_url
        else:
            url = cls._build_url(name="", namespace=namespace)
        resp = cls._client.get(url)
        return [cls.from_dict(item) for item in resp.json()["items"]]

    @classmethod
    def watch_list(cls, namespace=None, resource_version=None, allow_bookmarks=False):
        """Return a generator that yields WatchEvents of cls.
        If allowBookmarks is True, WatchBookmarks will also be yielded.
        It's recommended to use the Watcher class instead of calling this directly,
        since it handles reconnects and resource versions.
        """
        url = cls._watch_list_url(namespace)

        # We don't pass timeoutSeconds to the server, since our timeout is between each event,
        # while the server will apply the timeout as a maximum time serving the full request,
        # hanging up regardless of time between events. Let the server decide that timeout.
        params = {}
        if resource_version:
            # As per https://kubernetes.io/docs/reference/using-api/api-concepts/#semantics-for-watch
            # only resourceVersion is used for watch queries.
            params["resourceVersion"] = resource_version
            LOG.info("Restarting %s watch at resource version %s", cls.__name__, resource_version)
        if allow_bookmarks:
            params["allowWatchBookmarks"] = "true"

        try:
            # The timeout here appears to be per call to the poll (or similar) system call,
            # so each time data is received, the timeout will reset.
            resp = cls._client.get(url, stream=True, timeout=config.stream_timeout, params=params)
            for line in resp.iter_lines(chunk_size=None):
                event = cls._parse_watch_event(line) if line else None
                if event:
                    yield event
        except requests.ConnectionError as e:
            # ConnectionError is fairly generic, but check for ReadTimeoutError from urllib3.
            # If we get this, there were no events received for the timeout period, which might not be an error,
            # just a quiet period.
            underlying = e.args[0]
            if isinstance(underlying, urllib3.exceptions.ReadTimeoutError):
                LOG.info(
                    "Read timeout while waiting for new %s events.",
                    cls.__name__,
                )
                return
            raise

    @classmethod
    def _watch_list_url(cls, namespace):
        """Loads the optionally namespaced url from the class meta"""
        if namespace:
            if cls._meta.watch_list_url_template:
                url = cls._meta.watch_list_url_template.format(namespace=namespace)
            else:
                raise NotImplementedError(
                    "Cannot watch_list with namespace, no watch_list_url_template defined on class {}".format(cls)
                )
        else:
            url = cls._meta.watch_list_url
            if not url:
                raise NotImplementedError("Cannot watch_list, no watch_list_url defined on class {}".format(cls))
        return url

    @classmethod
    def _parse_watch_event(cls, line) -> Optional[WatchBaseEvent]:
        """
        Parse a line from the watch stream into a WatchEvent or WatchBookmark.
        Raises APIServerError if the line is an error event.
        """
        try:
            event_json = json.loads(line)
            if APIServerError.match(event_json):
                err = APIServerError(event_json["object"])
                LOG.warning(
                    "Received error event from API server: %s",
                    err,
                )
                raise err
            if WatchBookmark.match(event_json):
                LOG.debug("Received bookmark from API server: %s", event_json)
                event = WatchBookmark(event_json)
            else:
                LOG.debug("Received watch event from API server: %s", event_json)
                event = WatchEvent(event_json, cls)
            return event
        except TypeError:
            LOG.exception(
                "Unable to create instance of %s from watch event json, discarding event. event_json=%r",
                cls.__name__,
                event_json,
            )
        except ValueError:
            LOG.exception(
                "Unable to parse JSON on watch event, discarding event. Line: %r",
                line,
            )
        return None

    @classmethod
    def get(cls, name, namespace="default"):
        """Get from API server if it exists"""
        url = cls._build_url(name=name, namespace=namespace)
        resp = cls._client.get(url)
        instance = cls.from_dict(resp.json())
        return instance

    @classmethod
    def get_or_create(cls, **kwargs):
        """If exists, get from API, else create new instance"""
        try:
            metadata = kwargs.get("metadata")
            instance = cls.get(metadata.name, metadata.namespace)
            for field in cls._meta.fields:
                field.set(instance, kwargs)
            return instance
        except NotFound:
            return cls(new=True, **kwargs)

    @classmethod
    def delete(cls, name, namespace="default", **kwargs):
        """Delete the named resource"""
        url = cls._build_url(name=name, namespace=namespace)
        cls._client.delete(url, **kwargs)

    @classmethod
    def delete_list(cls, namespace="default", labels=None, delete_options=None, **kwargs):
        selector = cls._label_selector(labels)
        url = cls._build_url(name="", namespace=namespace)
        if delete_options:
            delete_options = delete_options.as_dict()

        cls._client.delete(url, body=delete_options, params={"labelSelector": selector}, **kwargs)

    def save(self):
        """Save to API server, either update if existing, or create if new"""
        if self._new:
            url = self._build_url(name="", namespace=self.metadata.namespace)
            resp = self._client.post(url, self.as_dict())
            self._new = False
        else:
            url = self._build_url(name=self.metadata.name, namespace=self.metadata.namespace)
            resp = self._client.put(url, self.as_dict())
        self.update_from_dict(resp.json())

    def save_status(self):
        """Save status to API server, always updating"""
        url = self._build_url(name=self.metadata.name, namespace=self.metadata.namespace) + "/status"
        resp = self._client.put(url, self.as_dict())
        self.update_from_dict(resp.json())

    @staticmethod
    def _label_selector(labels):
        """Build a labelSelector string from a collection of key/values. The parameter can be either
        a dict, or a list of (key, value) tuples (this allows for repeating a key).

        The keys/values are used to build the `labelSelector` parameter to the API,
        and supports all the operations of the API through the use of :py:class:`~k8s.base.LabelSelector`.

        Each key is a label name. The value defines which operation to perform.
        Operations that take a single string value:

            - :py:class:`~k8s.base.Equality`
            - :py:class:`~k8s.base.Inequality`

        A plain string will automatically be wrapped by :py:class:`~k8s.base.Equality` for compatability
        with older versions of this method.

        Operations that take a sequence of string values:

            - :py:class:`~k8s.base.In`
            - :py:class:`~k8s.base.NotIn`

        Operations that takes no value:

            - :py:class:`~k8s.base.Exists`
        """

        if hasattr(labels, "items"):
            labels = sorted(labels.items(), key=lambda kv: kv[0])

        return ",".join("{}{}".format(k, v if isinstance(v, LabelSelector) else Equality(v)) for k, v in labels)


class Model(metaclass=MetaModel):
    """A kubernetes Model object

    Contains fields for each attribute in the API specification, and methods for export/import.
    """

    def __init__(self, new=True, **kwargs):
        self._new = new
        self._values = {}
        kwarg_names = set(kwargs.keys())
        for field in self._meta.fields:
            kwarg_names.discard(field.name)
            field.set(self, kwargs)
            if field.type == SelfModel:
                field.type = self.__class__
                field.default_value_create_instance = False
        if kwarg_names:
            raise TypeError(
                "{}() got unexpected keyword-arguments: {}".format(self.__class__.__name__, ", ".join(kwarg_names))
            )
        if self._new:
            self._validate_fields()

    def _validate_fields(self):
        for field in self._meta.fields:
            if not field.is_valid(self):
                raise TypeError("Value of field {} is not valid on {}".format(field.name, self))

    def as_dict(self):
        if all(getattr(self, field.attr_name) == field.default_value for field in self._meta.fields):
            return None
        d = {}
        for field in self._meta.fields:
            value = field.dump(self)
            if value is not None:
                d[_api_name(field.name)] = value
        return d

    def merge(self, other):
        """
        `merge` sets each field in `self` to the value provided by `other`
        This is mostly equivalent to just replacing `self` with `other`,
        except read only fields in `self` are preserved.
        """
        for field in self._meta.fields:
            setattr(self, field.name, getattr(other, field.name))

    update = merge  # For backwards compatibility

    def update_from_dict(self, d):
        for field in self._meta.fields:
            field.load(self, d.get(_api_name(field.name)))
        self._validate_fields()

    @classmethod
    def from_dict(cls, d):
        instance = cls(new=False)
        instance.update_from_dict(d)
        return instance

    def __repr__(self):
        return "{}({})".format(
            self.__class__.__name__,
            ", ".join("{}={}".format(key, getattr(self, key)) for key in self._meta.field_names),
        )

    def __eq__(self, other):
        try:
            return self.as_dict() == other.as_dict()
        except AttributeError:
            return False


def _api_name(name):
    return name[1:] if name.startswith("_") else name


class WatchBaseEvent(ABC):
    """Abstract base class for Watch events.
    Contains the resource version of the event as property resource_version."""

    __slots__ = ("resource_version",)

    def __init__(self, event_json):
        self.resource_version = event_json["object"].get("metadata", {}).get("resourceVersion")

    def __eq__(self, other):
        return self.resource_version == other.resource_version

    def has_object(self):
        ...


class WatchEvent(WatchBaseEvent):
    ADDED = "ADDED"
    MODIFIED = "MODIFIED"
    DELETED = "DELETED"

    def __init__(self, event_json, cls):
        super(WatchEvent, self).__init__(event_json)
        self.type = event_json["type"]
        self.object = cls.from_dict(event_json["object"])

    def __repr__(self):
        return "{cls}(type={type}, object={object})".format(
            cls=self.__class__.__name__, type=self.type, object=self.object
        )

    def __eq__(self, other):
        return self.type == other.type and self.object == other.object

    def has_object(self):
        return True


class WatchBookmark(WatchBaseEvent):
    """Bookmark events, if enabled, are sent periodically by the API server.
    They only contain the resourceVersion of the event."""

    def __init__(self, event_json):
        super(WatchBookmark, self).__init__(event_json)

    @classmethod
    def match(cls, event_json):
        return event_json["type"] == "BOOKMARK"

    def has_object(self):
        return False


class LabelSelector(object):
    """Base for label select operations"""

    #: Operator used in selection query
    operator = None

    def __init__(self, value=""):
        self.value = value

    def __str__(self):
        return "{}{}".format(self.operator, self.value)


class Equality(LabelSelector):
    operator = "="


class Inequality(LabelSelector):
    operator = "!="


class LabelSetSelector(LabelSelector):
    def __str__(self):
        return " {} ({})".format(self.operator, ",".join(self.value))


class In(LabelSetSelector):
    operator = "in"


class NotIn(LabelSetSelector):
    operator = "notin"


class Exists(LabelSelector):
    def __init__(self):
        super(Exists, self).__init__("")

    def __str__(self):
        return ""


class SelfModel:
    """
    Use `SelfModel` as `Field.type` to set `Field.type` to the model the
    `Field` was defined in during model instantiation.

    This allows models to have fields with their own type.
    It is not possible to reference a class in its own attributes.

    Example:
    ```
    class MyModel(Model):
        submodel = Field(SelfModel) # submodel gets the type `MyModel`
    ```
    """

    pass


class APIServerError(Exception):
    """Raised when the API server returns an error event in the watch stream"""

    def __init__(self, api_error):
        self.api_error = api_error

    def __str__(self):
        code = self.api_error["code"]
        reason = self.api_error.get("reason", "(unset)")
        message = self.api_error.get("message", "(unset)")
        return f"{code}: reason={reason} message={message}"

    @classmethod
    def match(cls, event_json):
        return event_json["type"] == "ERROR" and event_json["object"].get("kind") == "Status"



#!/usr/bin/env python
# -*- coding: utf-8

# Copyright 2017-2020 The FIAAS Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import datetime

from .common import ObjectMeta, ObjectReference
from .job import JobTemplateSpec
from ..base import Model
from ..fields import Field, ListField, ReadOnlyField


class CronJobSpec(Model):
    concurrencyPolicy = Field(str)
    failedJobsHistoryLimit = Field(int)
    jobTemplate = Field(JobTemplateSpec)
    schedule = Field(str)
    startingDeadlineSeconds = Field(int)
    successfulJobsHistoryLimit = Field(int)
    suspend = Field(bool)


class CronJobStatus(Model):
    active = ListField(ObjectReference)
    lastScheduleTime = ReadOnlyField(datetime.datetime)


class CronJob(Model):
    class Meta:
        list_url = "/apis/batch/v1beta1/cronjobs"
        url_template = "/apis/batch/v1beta1/namespaces/{namespace}/cronjobs/{name}"

    metadata = Field(ObjectMeta)
    spec = Field(CronJobSpec)
    status = Field(CronJobStatus)



#!/usr/bin/env python
# -*- coding: utf-8

# Copyright 2017-2019 The FIAAS Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime

from .common import ObjectMeta
from ..base import Model, SelfModel
from ..fields import Field, ListField, JSONField


class ExternalDocumentation(Model):
    description = Field(str)
    url = Field(str)


class JSONSchemaProps(Model):
    ref = Field(str, name='$ref')
    schema = Field(str, name='$schema')
    additionalItems = Field(SelfModel, alt_type=bool)
    additionalProperties = Field(SelfModel, alt_type=bool)
    allOf = ListField(SelfModel)
    anyOf = ListField(SelfModel)
    default = JSONField()
    definitions = Field(dict)
    dependencies = Field(dict)
    description = Field(str)
    enum = JSONField()
    example = JSONField()
    exclusiveMaximum = Field(bool)
    exclusiveMinimum = Field(bool)
    externalDocs = Field(ExternalDocumentation)
    format = Field(str)
    id = Field(str)
    items = Field(SelfModel, alt_type=list)
    maxItems = Field(int)
    maxLength = Field(int)
    maxProperties = Field(int)
    maximum = Field(int, alt_type=float)
    minItems = Field(int)
    minLength = Field(int)
    minProperties = Field(int)
    minimum = Field(int, alt_type=float)
    multipleOf = Field(int, alt_type=float)
    _not = Field(SelfModel)
    nullable = Field(bool)
    oneOf = ListField(SelfModel)
    pattern = Field(str)
    patternProperties = Field(dict)
    properties = Field(dict)
    required = ListField(str)
    title = Field(str)
    type = Field(str)
    uniqueItems = Field(bool)
    x_kubernetes_embedded_resource = Field(bool, name='x-kubernetes-embedded-resource')
    x_kubernetes_int_or_string = Field(bool, name='x-kubernetes-int-or-string')
    x_kubernetes_list_map_keys = ListField(str, name='x-kubernetes-list-map-keys')
    x_kubernetes_list_type = Field(str, name="x-kubernetes-list-type")
    x_kubernetes_map_type = Field(str, name='x-kubernetes-map-type')
    x_kubernetes_preserve_unknown_fields = Field(bool, name="x-kubernetes-preserve-unknown-fields")


class JSONSchemaPropsStatusEnabled(Model):
    ref = Field(str, name='$ref')
    schema = Field(str, name='$schema')

    description = Field(str)
    example = JSONField()
    exclusiveMaximum = Field(bool)
    exclusiveMinimum = Field(bool)
    externalDocs = Field(ExternalDocumentation)
    format = Field(str)
    items = Field(SelfModel, alt_type=list)
    maxItems = Field(int)
    maxLength = Field(int)
    maximum = Field(int, alt_type=float)
    minItems = Field(int)
    minLength = Field(int)
    minimum = Field(int, alt_type=float)
    multipleOf = Field(int, alt_type=float)
    pattern = Field(str)
    properties = Field(dict)
    required = ListField(str)
    title = Field(str)
    type = Field(str)


class ServiceReference(Model):
    name = Field(str)
    namespace = Field(str)
    path = Field(str)
    port = Field(int)


class WebhookClientConfig(Model):
    caBundle = Field(str)
    service = Field(ServiceReference)
    url = Field(str)


class WebhookConversion(Model):
    clientConfig = Field(WebhookClientConfig)
    conversionReviewVersions = ListField(str)


class CustomResourceColumnDefinition(Model):
    description = Field(str)
    format = Field(str)
    jsonPath = Field(str)
    name = Field(str)
    priority = Field(int)
    type = Field(str)


class CustomResourceConversion(Model):
    strategy = Field(str)
    webhook = Field(WebhookConversion)


class CustomResourceDefinitionNames(Model):
    categories = ListField(str)
    kind = Field(str)
    listKind = Field(str)
    plural = Field(str)
    shortNames = ListField(str)
    singular = Field(str)


class CustomResourceValidation(Model):
    openAPIV3Schema = Field(JSONSchemaProps, alt_type=JSONSchemaPropsStatusEnabled)


class CustomResourceSubresourceScale(Model):
    labelSelectorPath = Field(str)
    specReplicasPath = Field(str)
    statusReplicasPath = Field(str)


class CustomResourceSubresourceStatusEnabled(Model):
    def as_dict(self):
        return {}


class CustomResourceSubresourceStatusDisabled(Model):
    pass


class CustomResourceSubresources(Model):
    scale = Field(CustomResourceSubresourceScale)
    status = Field(CustomResourceSubresourceStatusDisabled, alt_type=CustomResourceSubresourceStatusEnabled)


class CustomResourceDefinitionVersion(Model):
    additionalPrinterColumns = ListField(CustomResourceColumnDefinition)
    deprecated = Field(bool)
    deprecationWarning = Field(str)
    name = Field(str)
    schema = Field(CustomResourceValidation)
    served = Field(bool)
    storage = Field(bool)
    subresources = Field(CustomResourceSubresources)


class CustomResourceDefinitionSpec(Model):
    conversion = Field(CustomResourceConversion)
    group = Field(str)
    names = Field(CustomResourceDefinitionNames)
    preserveUnknownFields = Field(bool)
    scope = Field(str)
    versions = ListField(CustomResourceDefinitionVersion)


class CustomResourceDefinitionCondition(Model):
    lastTransitionTime = Field(datetime.datetime)
    message = Field(str)
    reason = Field(str)
    status = Field(str)
    type = Field(str)


class CustomResourceDefinitionStatus(Model):
    acceptedNames = Field(CustomResourceDefinitionNames)
    conditions = ListField(CustomResourceDefinitionCondition)
    storedVersions = ListField(str)


class CustomResourceDefinition(Model):
    class Meta:
        url_template = "/apis/apiextensions.k8s.io/v1/customresourcedefinitions/{name}"
        list_url = "/apis/apiextensions.k8s.io/v1/customresourcedefinitions"
        watch_list_url = "/apis/apiextensions.k8s.io/v1/customresourcedefinitions?watch=true"

    metadata = Field(ObjectMeta)
    spec = Field(CustomResourceDefinitionSpec)
    status = Field(CustomResourceDefinitionStatus)



#!/usr/bin/env python
# -*- coding: utf-8

# Copyright 2017-2019 The FIAAS Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import cachetools

from .base import APIServerError, WatchEvent

DEFAULT_CAPACITY = 1000


class Watcher(object):
    """Higher-level interface to watch for changes in objects

    The low-level :py:meth:`~.watch_list` method will stop when the API-server drops the connection.
    When reconnecting using that method, the API-server will send a list of :py:const:`~k8s.base.WatchEvent.ADDED`
    events for all objects, even if they have been seen before.

    The Watcher will hide this complexity for you, and make sure to reconnect when the
    connection drops, and skip events that have already been seen.
    It additionally uses bookmarks to avoid the increased load that might be caused by reconnecting.

    :param Model model: The model class to watch
    :param int capacity: How many seen objects to keep track of
    """

    def __init__(self, model, capacity=DEFAULT_CAPACITY):
        self._seen = cachetools.LRUCache(capacity)
        self._model = model
        self._run_forever = True

    def watch(self, namespace=None):
        """Watch for events

        :param str namespace: the namespace to watch for events in. The default (None) results in
            watching for events in all namespaces.
        :return: a generator that yields :py:class:`~.WatchEvent` objects not seen before
        """
        # last_seen_resource_version is used to resume the watch from the last seen event.
        # Only used on reconnects, the first call to watch does a quorum read.
        last_seen_resource_version = None
        while self._run_forever:
            try:
                for event in self._model.watch_list(
                    namespace=namespace, resource_version=last_seen_resource_version, allow_bookmarks=True
                ):
                    last_seen_resource_version = event.resource_version
                    if self._should_yield(event):
                        yield event
            except APIServerError as e:
                # A 410 response indicates our resourceVersion is too old, and we need to do a new quorum read.
                if e.api_error["code"] == 410:
                    last_seen_resource_version = None
                else:
                    raise

    def _should_yield(self, event) -> bool:
        """Check if this is a new event, and if so, mark it as seen"""
        if not event.has_object():
            return False
        o = event.object
        key = (o.metadata.name, o.metadata.namespace)
        if self._seen.get(key) == o.metadata.resourceVersion and event.type != WatchEvent.DELETED:
            return False
        self._seen[key] = o.metadata.resourceVersion
        return True



#!/usr/bin/env python
# -*- coding: utf-8

# Copyright 2017-2019 The FIAAS Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import mock
import pytest

from k8s.client import NotFound
from k8s.models.common import ObjectMeta
from k8s.models.role import Role, PolicyRule

NAME = "my-role"
NAMESPACE = "my-namespace"


@pytest.mark.usefixtures("k8s_config")
class TestRole(object):
    def test_created_if_not_exists(self, post, api_get):
        api_get.side_effect = NotFound()
        role = _create_default_role()
        call_params = role.as_dict()
        post.return_value.json.return_value = call_params

        assert role._new
        role.save()
        assert not role._new

        pytest.helpers.assert_any_call(post, _uri(NAMESPACE), call_params)

    def test_updated_if_exists(self, get, put):
        mock_response = _create_mock_response()
        get.return_value = mock_response
        role = _create_default_role()

        from_api = Role.get_or_create(
            metadata=role.metadata,
            rules=role.rules,
        )
        assert not from_api._new
        assert from_api.rules == role.rules

    def test_deleted(self, delete):
        Role.delete(NAME, namespace=NAMESPACE)
        pytest.helpers.assert_any_call(delete, _uri(NAMESPACE, NAME))


def _create_mock_response():
    mock_response = mock.Mock()
    mock_response.json.return_value = {
        "apiVersion": "rbac.authorization.k8s.io/v1",
        "kind": "Role",
        "metadata": {
            "creationTimestamp": "2017-09-08T13:37:00Z",
            "generation": 1,
            "labels": {
                "test": "true"
            },
            "name": NAME,
            "namespace": NAMESPACE,
            "resourceVersion": "42",
            "selfLink": _uri(NAMESPACE, NAME),
            "uid": "d8f1ba26-b182-11e6-a364-fa163ea2a9c4"
        },
        "rules": [
            {
                "apiGroups": ["fiaas.schibsted.io"],
                "resources": ["applications", "application-statuses"],
                "verbs": ["get", "list", "watch"],
            },
        ],
    }
    return mock_response


def _create_default_role():
    object_meta = ObjectMeta(name=NAME, namespace=NAMESPACE, labels={"test": "true"})
    policy_rules = [
        PolicyRule(
            apiGroups=["fiaas.schibsted.io"],
            resources=["applications", "application-statuses"],
            verbs=["get", "list", "watch"],
            resourceNames=[],
            nonResourceURLs=[],
        )
    ]
    return Role(metadata=object_meta, rules=policy_rules)


def _uri(namespace, name=""):
    uri = "/apis/rbac.authorization.k8s.io/v1/namespaces/{namespace}/roles/{name}"
    return uri.format(name=name, namespace=namespace)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2017-2019 The FIAAS Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
import pytest
import requests
import requests.packages.urllib3 as urllib3

from k8s.base import APIServerError, Equality, Exists, Field, In, Inequality, Model, NotIn, WatchBookmark, WatchEvent
from k8s.models.common import DeleteOptions, Preconditions


class Example(Model):
    class Meta:
        url_template = '/example'
        watch_list_url = '/watch/example'

    value = Field(int)


class TestWatchEvent(object):
    def test_watch_event_added(self):
        watch_event = WatchEvent({"type": "ADDED", "object": {"value": 42}}, Example)
        assert watch_event.type == WatchEvent.ADDED
        assert watch_event.object == Example(value=42)

    def test_watch_event_modified(self):
        watch_event = WatchEvent({"type": "MODIFIED", "object": {"value": 42}}, Example)
        assert watch_event.type == WatchEvent.MODIFIED
        assert watch_event.object == Example(value=42)

    def test_watch_event_deleted(self):
        watch_event = WatchEvent({"type": "DELETED", "object": {"value": 42}}, Example)
        assert watch_event.type == WatchEvent.DELETED
        assert watch_event.object == Example(value=42)


class TestFind(object):
    @pytest.fixture
    def client(self):
        with mock.patch.object(Example, "_client") as m:
            yield m

    def test_find_by_name(self, client):
        Example.find("app_name")
        client.get.assert_called_once_with("/example", params={"labelSelector": "app=app_name"})

    @pytest.mark.parametrize("value, selector", (
        (Equality("my_value"), "my_key=my_value"),
        (Inequality("my_value"), "my_key!=my_value"),
        (In(("value1", "value2")), "my_key in (value1,value2)"),
        (NotIn(("value1", "value2")), "my_key notin (value1,value2)"),
        (Exists(), "my_key"),
        ("my_unwrapped_value", "my_key=my_unwrapped_value"),
    ))
    def test_find_by_selectors(self, client, value, selector):
        Example.find(labels={"my_key": value})
        client.get.assert_called_once_with("/example", params={"labelSelector": selector})

    def test_repeated_keys_in_label_selector(self, client):
        labels = [
            ("foo", Inequality("bar")),
            ("foo", Exists())
        ]
        Example.find(labels=labels)

        expected_selector = "foo!=bar,foo"

        client.get.assert_called_once_with("/example", params={"labelSelector": expected_selector})


class TestDeleteList(object):

    @pytest.fixture
    def client(self):
        with mock.patch.object(Example, "_client") as m:
            yield m

    def test_delete_list(self, client):
        Example.delete_list(labels={"foo": Equality("bar"), "cat": Inequality("dog")})

        client.delete.assert_called_once_with("/example", body=None, params={"labelSelector": "cat!=dog,foo=bar"})

    def test_delete_with_options(self, client):
        opts = DeleteOptions(
            apiVersion="foo/v1",
            dryRun=[],
            gracePeriodSeconds=30,
            preconditions=Preconditions(uid="1234", resourceVersion="12"),
            propagationPolicy="Foreground"
        )
        Example.delete_list(labels={"foo": "bar"}, delete_options=opts)

        expected_body = {
            "apiVersion": "foo/v1",
            "dryRun": [],
            "gracePeriodSeconds": 30,
            "preconditions": {
                "uid": "1234",
                "resourceVersion": "12"
            },
            "propagationPolicy": "Foreground"
        }
        client.delete.assert_called_once_with("/example", params={"labelSelector": "foo=bar"}, body=expected_body)


class TestWatchList(object):
    @pytest.fixture
    def client(self):
        with mock.patch.object(Example, "_client") as m:
            yield m

    def test_watch_list(self, client):
        client.get.return_value.iter_lines.return_value = [
            '{"type": "ADDED", "object": {"value": 1}}',
        ]
        gen = Example.watch_list()
        assert next(gen) == WatchEvent({"type": "ADDED", "object": {"value": 1}}, Example)
        client.get.assert_called_once_with("/watch/example", stream=True, timeout=270, params={})
        assert list(gen) == []

    def test_watch_list_with_timeout(self, client):
        client.get.return_value.iter_lines.return_value.__getitem__.side_effect = [
            '{"type": "ADDED", "object": {"value": 1}}',
            requests.ConnectionError(urllib3.exceptions.ReadTimeoutError("", "", "")),
            '{"type": "MODIFIED", "object": {"value": 2}}',  # Not reached
        ]
        # Seal to avoid __iter__ being used instead of __getitem__
        mock.seal(client)
        gen = Example.watch_list()
        assert next(gen) == WatchEvent({"type": "ADDED", "object": {"value": 1}}, Example)
        assert list(gen) == []
        assert client.get.return_value.iter_lines.return_value.__getitem__.call_count == 2
        client.get.assert_called_once_with("/watch/example", stream=True, timeout=270, params={})

    def test_watch_list_api_error(self, client):
        client.get.return_value.iter_lines.return_value = [
            '{"type": "ERROR", "object": {"kind":"Status", "code": 500, "message": "Internal Server Error"}}',
        ]
        gen = Example.watch_list()
        with pytest.raises(APIServerError, match="Internal Server Error"):
            next(gen)

    def test_watch_list_bookmark(self, client):
        client.get.return_value.iter_lines.return_value = [
            '{"type":"BOOKMARK", "object":{"metadata":{"resourceVersion": 4712}}}',
        ]
        gen = Example.watch_list(resource_version=4711, allow_bookmarks=True)
        assert next(gen) == WatchBookmark({"type": "BOOKMARK", "object": {"metadata": {"resourceVersion": 4712}}})
        assert list(gen) == []
        client.get.assert_called_once_with(
            "/watch/example", stream=True, timeout=270, params={"resourceVersion": 4711, "allowWatchBookmarks": "true"}
        )
